/*
 * mod_injection http://pmade.org/pjones/software/mod_injection/
 *
 * Copyright (C) 2002-2003 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_request.h"
#include "ap_config.h"
#include "apr_buckets.h"
#include "apr_strings.h"
#include "apr_lib.h"
#include "apr_tables.h"
#include "apr_fnmatch.h"
#include "util_filter.h"
#include <string.h>

/* ######################################################################## */
static const char const_filter_name[] = "INJECTION";
module AP_MODULE_DECLARE_DATA injection_module;
/* ######################################################################## */
typedef struct injection_ctx_t {
    enum parse_state_t {
	parse_find_start,
	parse_match_string,
	parse_find_end
    } parse_state;

    const char *match_pos;
} injection_ctx_t;
/* ######################################################################## */
typedef struct injection_cfg_t {
    const char *inject_string;
    const char *inject_match;

    enum string_type_t {
	string_notset,
	string_raw,
	string_uri
    } string_type;

    enum match_type_t {
	match_tag,
	match_string,
    } match_type;

    apr_array_header_t *content_types;
} injection_cfg_t;
/* ######################################################################## */
static void* injection_create_config (apr_pool_t *p) {
    injection_cfg_t *cfg = apr_pcalloc(p, sizeof(*cfg));

    cfg->inject_string = 0;
    cfg->inject_match = "body";
    cfg->string_type = string_notset;
    cfg->match_type = match_tag;
    cfg->content_types = apr_array_make(p, 0, sizeof(char*));

    return cfg;
}
/* ######################################################################## */
static void* injection_create_server_config (apr_pool_t *p, server_rec *s) {
    return injection_create_config(p);
}
/* ######################################################################## */
static void* injection_create_dir_config (apr_pool_t *p, char *dirspec) {
    return injection_create_config(p);
}
/* ######################################################################## */
static const char* cfg_inject_tag (cmd_parms *parms, void *mconfig, const char *w) {
    injection_cfg_t *cfg = mconfig;
    apr_size_t strsize = strlen(w);
    char *tag;

    if (strcmp(cfg->inject_match, "body") != 0) {
	return "InjectTag or InjectAfter already used";
    }

    if (*w == '<') { ++w; --strsize; }
    if (w[strsize-1] == '>') --strsize;

    tag = apr_pstrmemdup(parms->pool, w, strsize);
    ap_str_tolower(tag);

    cfg->inject_match = tag;
    cfg->match_type = match_tag;

    return NULL;
}
/* ######################################################################## */
static const char* cfg_inject_after (cmd_parms *parms, void *mconfig, const char *w) {
    injection_cfg_t *cfg = mconfig;

    if (strcmp(cfg->inject_match, "body") != 0) {
	return "InjectTag or InjectAfter already used";
    }

    cfg->inject_match = apr_pstrdup(parms->pool, w);
    cfg->match_type = match_string;

    return NULL;
}
/* ######################################################################## */
static const char* cfg_inject_string (cmd_parms *parms, void *mconfig, const char *w) {
    injection_cfg_t *cfg = mconfig;

    if (cfg->inject_string) {
	return "injection method already set";
    }

    cfg->inject_string = apr_pstrdup(parms->pool, w);
    cfg->string_type = string_raw;

    return NULL;
}
/* ######################################################################## */
static const char* cfg_inject_uri (cmd_parms *parms, void *mconfig, const char *w) {
    injection_cfg_t *cfg = mconfig;

    if (cfg->inject_string) {
	return "injection method already set";
    }

    cfg->inject_string = apr_pstrdup(parms->pool, w);
    cfg->string_type = string_uri;

    return NULL;
}
/* ######################################################################## */
static const char* cfg_inject_type (cmd_parms *parms, void *mconfig, const char *w) {
    injection_cfg_t *cfg = mconfig;

    char **new_element = apr_array_push(cfg->content_types);
    *new_element = apr_pstrdup(parms->pool, w);

    return NULL;
}
/* ######################################################################## */
static apr_status_t  handle_injection (ap_filter_t *f, apr_bucket_brigade **bb, apr_bucket *e, injection_cfg_t *cfg) {
    request_rec *rr = NULL;
    apr_status_t rc;
    apr_bucket_brigade *new_bb;
    apr_bucket *p;
    apr_size_t len;

    switch (cfg->string_type) {
	case string_notset:
	    break;
	case string_raw:
	    len = strlen(cfg->inject_string);
	    p = apr_bucket_pool_create(cfg->inject_string, len, f->r->pool, f->c->bucket_alloc);
	    APR_BUCKET_INSERT_AFTER(e, p);
	    break;
	case string_uri:
	    new_bb = apr_brigade_split(*bb, APR_BUCKET_NEXT(e));
	    if ( (rc = ap_pass_brigade(f->next, *bb)) != APR_SUCCESS) return rc;
	    *bb = new_bb;
	    rr = ap_sub_req_lookup_uri(cfg->inject_string, f->r, f->next);
	    if (rr->status != HTTP_OK || ap_run_sub_req(rr)) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, rr, "failed to inject URI: %s", cfg->inject_string);
		return AP_FILTER_ERROR;
	    }
	    break;
    }

    if (rr) ap_destroy_sub_req(rr);
    ap_remove_output_filter(f);
    return ap_pass_brigade(f->next, *bb);
}
/* ######################################################################## */
static apr_status_t injection_filter (ap_filter_t *f, apr_bucket_brigade *bb) {
    injection_ctx_t *ctx = f->ctx;
    injection_cfg_t *cfg;
    apr_bucket *e;
    const char *data;
    apr_size_t len, i;

    /*
     * There are various reasons that we should not run.
     */
    if (f->r->main || (f->r->status != HTTP_OK && f->r->status != HTTP_NON_AUTHORITATIVE) || !f->r->content_type) {
	ap_remove_output_filter(f);
	return ap_pass_brigade(f->next, bb);
    }

    /* 
     * Load our configuration info
     */
    cfg = ap_get_module_config(f->r->per_dir_config, &injection_module);

    if (!cfg || cfg->string_type == string_notset) {
	/* try the server config */
	cfg = ap_get_module_config(f->r->server->module_config, &injection_module);
    }

    if (!cfg || cfg->string_type == string_notset) {
	/* still no go */
	ap_remove_output_filter(f);
	return ap_pass_brigade(f->next, bb);
    }

    /*
     * If this is our first time being called we need to create our context
     * variable after doing some basic checking
     */
    if (!ctx) {
	/*
	 * Default is to only inject for text/html only
	 */
	if (cfg->content_types->nelts == 0) {
	   if (strcmp(f->r->content_type, "text/html")) {
	       ap_remove_output_filter(f);
	       return ap_pass_brigade(f->next, bb);
	   }
	} else {
	    char **element = (char**)cfg->content_types->elts;
	    int allow=0;
	    int i=0;

	    /*
	     * Allow MIME types to have shell file glob patterns
	     */
	    for (; i<cfg->content_types->nelts; ++i) {
		if (apr_fnmatch(element[i], f->r->content_type, APR_FNM_PATHNAME) == APR_SUCCESS) {
		    allow = 1;
		    break;
		}
	    }

	    if (!allow) {
		ap_remove_output_filter(f);
		return ap_pass_brigade(f->next, bb);
	    }
	}

	ctx = f->ctx = apr_pcalloc(f->r->pool, sizeof(*ctx));
	ctx->parse_state = parse_find_start;
	ctx->match_pos = cfg->inject_match;

	/* 
	 * delete the content-length header because we may change the
	 * content and thus its length
	 */
	apr_table_unset(f->r->headers_out, "Content-Length");
    }

    for (e = APR_BRIGADE_FIRST(bb);
         e != APR_BRIGADE_SENTINEL(bb);
         e = APR_BUCKET_NEXT(e)) {
	/*
	 * scan each bucket looking for a body. pass each brigade to the
	 * next filter as we are done with them. if we find the body tag, we
	 * can remove ourself as a filter and pass the brigade to the next
	 * filter.
	 */

	if (APR_BUCKET_IS_EOS(e)) {
	    /* we only got here if we did not inject something */
	    return ap_pass_brigade(f->next, bb);
	}

	if (APR_BUCKET_IS_FLUSH(e)) {
	    continue;
	}

	apr_bucket_read(e, &data, &len, APR_BLOCK_READ);
	for (i=0; i<len; ++i) {
	    switch (ctx->parse_state) {
		case parse_find_start:
		    switch (cfg->match_type) {
			case match_tag:
			    if (data[i] == '<') ctx->parse_state = parse_match_string;
			    break;
			case match_string:
			    if (data[i] == *(ctx->match_pos)) {
				++(ctx->match_pos);
				ctx->parse_state = parse_match_string;
			    }
			    break;
		    }
		    break;

		case parse_match_string:
		    switch (cfg->match_type) {
			case match_tag:
			    if (apr_tolower(data[i]) != *(ctx->match_pos)) {
				ctx->match_pos = cfg->inject_match;
				ctx->parse_state = parse_find_start;
				break;
			    }
			    ++(ctx->match_pos);
			    if (*(ctx->match_pos) == '\0') {
				ctx->parse_state = parse_find_end;
				break;
			    }
			    break;
			case match_string:
			    if (*(ctx->match_pos) == '\0') {
				apr_bucket_split(e, i);
				return handle_injection(f, &bb, e, cfg);
				break;
			    }
			    if (data[i] != *(ctx->match_pos)) {
				ctx->parse_state = parse_find_start;
				ctx->match_pos = cfg->inject_match;
				break;
			    }
			    ++(ctx->match_pos);
		    }
		    break;

		case parse_find_end:
		    /* only match_type == match_tag here */
		    if (*(ctx->match_pos) == '\0') {
			if (!apr_isspace(data[i]) && data[i] != '/' && data[i] != '>') {
			    /* wrong tag */
			    ctx->match_pos = cfg->inject_match;
			    ctx->parse_state = parse_find_start;
			    break;
			}

			ctx->match_pos = cfg->inject_match;
		    }
		    if (data[i] == '>') {
			apr_bucket_split(e, i+1);
			return handle_injection(f, &bb, e, cfg);
		    }
		    break;
	    }
	}
    }

    return ap_pass_brigade(f->next, bb);
}
/* ######################################################################## */
static int injection_init_handler(apr_pool_t *p, apr_pool_t *plog,
                                apr_pool_t *ptemp, server_rec *s)
{
    ap_add_version_component(p, "mod_injection/0.3.1");
    return OK;
}
/* ######################################################################## */
static void injection_register_hooks(apr_pool_t *p) {
    ap_hook_post_config(injection_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_register_output_filter(const_filter_name, injection_filter, NULL, AP_FTYPE_CONTENT_SET);
}
/* ######################################################################## */
static const command_rec injection_cmds[] = {
    AP_INIT_TAKE1("InjectTag",    cfg_inject_tag,    NULL, OR_OPTIONS, "Inject after the given tag name"),
    AP_INIT_TAKE1("InjectAfter",  cfg_inject_after,  NULL, OR_OPTIONS, "Inject after the given string"),
    AP_INIT_TAKE1("InjectString", cfg_inject_string, NULL, OR_OPTIONS, "Inject the given string after <body>"),
    AP_INIT_TAKE1("InjectURI",    cfg_inject_uri,    NULL, OR_OPTIONS, "Inject the given URI after <body>"),
    AP_INIT_TAKE1("InjectType",   cfg_inject_type,   NULL, OR_OPTIONS, "Set the allowed MIME type for Injection"),
    {NULL}
};
/* ######################################################################## */
module AP_MODULE_DECLARE_DATA injection_module = {
    STANDARD20_MODULE_STUFF, 
    injection_create_dir_config,
    NULL,                  
    injection_create_server_config,
    NULL,                  
    injection_cmds,
    injection_register_hooks
};
